@@ -6,21 +6,31 @@ import threading
 import time
 import types
 import warnings
+import weakref
 from contextlib import contextmanager, ExitStack
 from math import inf
 from textwrap import dedent
+import gc
 
 import attr
 import outcome
 import sniffio
 import pytest
-from async_generator import async_generator
 
-from .tutil import slow, check_sequence_matches, gc_collect_harder
+from .tutil import (
+    slow,
+    check_sequence_matches,
+    gc_collect_harder,
+    ignore_coroutine_never_awaited_warnings,
+    buggy_pypy_asyncgens,
+    restore_unraisablehook,
+    create_asyncio_future_in_new_loop,
+)
+
 from ... import _core
+from .._run import DEADLINE_HEAP_MIN_PRUNE_THRESHOLD
 from ..._threads import to_thread_run_sync
 from ..._timeouts import sleep, fail_after
-from ..._util import aiter_compat
 from ...testing import (
     wait_all_tasks_blocked,
     Sequencer,
@@ -35,26 +45,6 @@ async def sleep_forever():
     return await _core.wait_task_rescheduled(lambda _: _core.Abort.SUCCEEDED)
 
 
-# Some of our tests need to leak coroutines, and thus trigger the
-# "RuntimeWarning: coroutine '...' was never awaited" message. This context
-# manager should be used anywhere this happens to hide those messages, because
-# (a) when expected they're clutter, (b) on CPython 3.5.x where x < 3, this
-# warning can trigger a segfault if we run with warnings turned into errors:
-#   https://bugs.python.org/issue27811
-@contextmanager
-def ignore_coroutine_never_awaited_warnings():
-    with warnings.catch_warnings():
-        warnings.filterwarnings(
-            "ignore", message="coroutine '.*' was never awaited"
-        )
-        try:
-            yield
-        finally:
-            # Make sure to trigger any coroutine __del__ methods now, before
-            # we leave the context manager.
-            gc_collect_harder()
-
-
 def test_basic():
     async def trivial(x):
         return x
@@ -146,8 +136,7 @@ async def test_basic_interleave():
         nursery.start_soon(looper, "b", record)
 
     check_sequence_matches(
-        record,
-        [{("a", 0), ("b", 0)}, {("a", 1), ("b", 1)}, {("a", 2), ("b", 2)}]
+        record, [{("a", 0), ("b", 0)}, {("a", 1), ("b", 1)}, {("a", 2), ("b", 2)}]
     )
 
 
@@ -191,8 +180,10 @@ def test_main_and_task_both_crash():
     with pytest.raises(_core.MultiError) as excinfo:
         _core.run(main)
     print(excinfo.value)
-    assert {type(exc)
-            for exc in excinfo.value.exceptions} == {ValueError, KeyError}
+    assert {type(exc) for exc in excinfo.value.exceptions} == {
+        ValueError,
+        KeyError,
+    }
 
 
 def test_two_child_crashes():
@@ -206,8 +197,10 @@ def test_two_child_crashes():
 
     with pytest.raises(_core.MultiError) as excinfo:
         _core.run(main)
-    assert {type(exc)
-            for exc in excinfo.value.exceptions} == {ValueError, KeyError}
+    assert {type(exc) for exc in excinfo.value.exceptions} == {
+        ValueError,
+        KeyError,
+    }
 
 
 async def test_child_crash_wakes_parent():
@@ -285,7 +278,7 @@ async def test_current_task():
 
 async def test_root_task():
     root = _core.current_root_task()
-    assert root.parent_nursery is None
+    assert root.parent_nursery is root.eventual_parent_nursery is None
 
 
 def test_out_of_context():
@@ -347,202 +340,6 @@ async def test_current_statistics(mock_clock):
     assert stats.seconds_to_next_deadline == inf
 
 
-@attr.s(eq=False, hash=False)
-class TaskRecorder:
-    record = attr.ib(factory=list)
-
-    def before_run(self):
-        self.record.append(("before_run",))
-
-    def task_scheduled(self, task):
-        self.record.append(("schedule", task))
-
-    def before_task_step(self, task):
-        assert task is _core.current_task()
-        self.record.append(("before", task))
-
-    def after_task_step(self, task):
-        assert task is _core.current_task()
-        self.record.append(("after", task))
-
-    def after_run(self):
-        self.record.append(("after_run",))
-
-    def filter_tasks(self, tasks):
-        for item in self.record:
-            if item[0] in ("schedule", "before", "after") and item[1] in tasks:
-                yield item
-            if item[0] in ("before_run", "after_run"):
-                yield item
-
-
-def test_instruments(recwarn):
-    r1 = TaskRecorder()
-    r2 = TaskRecorder()
-    r3 = TaskRecorder()
-
-    task = None
-
-    # We use a child task for this, because the main task does some extra
-    # bookkeeping stuff that can leak into the instrument results, and we
-    # don't want to deal with it.
-    async def task_fn():
-        nonlocal task
-        task = _core.current_task()
-
-        for _ in range(4):
-            await _core.checkpoint()
-        # replace r2 with r3, to test that we can manipulate them as we go
-        _core.remove_instrument(r2)
-        with pytest.raises(KeyError):
-            _core.remove_instrument(r2)
-        # add is idempotent
-        _core.add_instrument(r3)
-        _core.add_instrument(r3)
-        for _ in range(1):
-            await _core.checkpoint()
-
-    async def main():
-        async with _core.open_nursery() as nursery:
-            nursery.start_soon(task_fn)
-
-    _core.run(main, instruments=[r1, r2])
-
-    # It sleeps 5 times, so it runs 6 times.  Note that checkpoint()
-    # reschedules the task immediately upon yielding, before the
-    # after_task_step event fires.
-    expected = (
-        [("before_run",), ("schedule", task)] +
-        [("before", task), ("schedule", task), ("after", task)] * 5 +
-        [("before", task), ("after", task), ("after_run",)]
-    )
-    assert len(r1.record) > len(r2.record) > len(r3.record)
-    assert r1.record == r2.record + r3.record
-    assert list(r1.filter_tasks([task])) == expected
-
-
-def test_instruments_interleave():
-    tasks = {}
-
-    async def two_step1():
-        tasks["t1"] = _core.current_task()
-        await _core.checkpoint()
-
-    async def two_step2():
-        tasks["t2"] = _core.current_task()
-        await _core.checkpoint()
-
-    async def main():
-        async with _core.open_nursery() as nursery:
-            nursery.start_soon(two_step1)
-            nursery.start_soon(two_step2)
-
-    r = TaskRecorder()
-    _core.run(main, instruments=[r])
-
-    expected = [
-        ("before_run",),
-        ("schedule", tasks["t1"]),
-        ("schedule", tasks["t2"]),
-        {
-            ("before", tasks["t1"]),
-            ("schedule", tasks["t1"]),
-            ("after", tasks["t1"]),
-            ("before", tasks["t2"]),
-            ("schedule", tasks["t2"]),
-            ("after", tasks["t2"])
-        },
-        {
-            ("before", tasks["t1"]),
-            ("after", tasks["t1"]),
-            ("before", tasks["t2"]),
-            ("after", tasks["t2"])
-        },
-        ("after_run",),
-    ]  # yapf: disable
-    print(list(r.filter_tasks(tasks.values())))
-    check_sequence_matches(list(r.filter_tasks(tasks.values())), expected)
-
-
-def test_null_instrument():
-    # undefined instrument methods are skipped
-    class NullInstrument:
-        pass
-
-    async def main():
-        await _core.checkpoint()
-
-    _core.run(main, instruments=[NullInstrument()])
-
-
-def test_instrument_before_after_run():
-    record = []
-
-    class BeforeAfterRun:
-        def before_run(self):
-            record.append("before_run")
-
-        def after_run(self):
-            record.append("after_run")
-
-    async def main():
-        pass
-
-    _core.run(main, instruments=[BeforeAfterRun()])
-    assert record == ["before_run", "after_run"]
-
-
-def test_instrument_task_spawn_exit():
-    record = []
-
-    class SpawnExitRecorder:
-        def task_spawned(self, task):
-            record.append(("spawned", task))
-
-        def task_exited(self, task):
-            record.append(("exited", task))
-
-    async def main():
-        return _core.current_task()
-
-    main_task = _core.run(main, instruments=[SpawnExitRecorder()])
-    assert ("spawned", main_task) in record
-    assert ("exited", main_task) in record
-
-
-# This test also tests having a crash before the initial task is even spawned,
-# which is very difficult to handle.
-def test_instruments_crash(caplog):
-    record = []
-
-    class BrokenInstrument:
-        def task_scheduled(self, task):
-            record.append("scheduled")
-            raise ValueError("oops")
-
-        def close(self):
-            # Shouldn't be called -- tests that the instrument disabling logic
-            # works right.
-            record.append("closed")  # pragma: no cover
-
-    async def main():
-        record.append("main ran")
-        return _core.current_task()
-
-    r = TaskRecorder()
-    main_task = _core.run(main, instruments=[r, BrokenInstrument()])
-    assert record == ["scheduled", "main ran"]
-    # the TaskRecorder kept going throughout, even though the BrokenInstrument
-    # was disabled
-    assert ("after", main_task) in r.record
-    assert ("after_run",) in r.record
-    # And we got a log message
-    exc_type, exc_value, exc_traceback = caplog.records[0].exc_info
-    assert exc_type is ValueError
-    assert str(exc_value) == "oops"
-    assert "Instrument has been disabled" in caplog.records[0].message
-
-
 async def test_cancel_scope_repr(mock_clock):
     scope = _core.CancelScope()
     assert "unbound" in repr(scope)
@@ -650,7 +447,7 @@ async def test_cancel_scope_multierror_filtering():
     except AssertionError:  # pragma: no cover
         raise
     except BaseException as exc:
-        # This is ouside the outer scope, so all the Cancelled
+        # This is outside the outer scope, so all the Cancelled
         # exceptions should have been absorbed, leaving just a regular
         # KeyError from crasher()
         assert type(exc) is KeyError
@@ -805,13 +602,6 @@ async def test_basic_timeout(mock_clock):
             await _core.checkpoint()
 
 
-@pytest.mark.filterwarnings(
-    "ignore:.*trio.open_cancel_scope:trio.TrioDeprecationWarning"
-)
-async def test_cancel_scope_deprecated(recwarn):
-    assert isinstance(_core.open_cancel_scope(), _core.CancelScope)
-
-
 async def test_cancel_scope_nesting():
     # Nested scopes: if two triggering at once, the outer one wins
     with _core.CancelScope() as scope1:
@@ -987,9 +777,7 @@ async def test_cancel_scope_misnesting():
         for exc in exc_info.value.__context__.exceptions:
             assert isinstance(exc, RuntimeError)
             assert "closed before the task exited" in str(exc)
-            cancelled_in_context |= isinstance(
-                exc.__context__, _core.Cancelled
-            )
+            cancelled_in_context |= isinstance(exc.__context__, _core.Cancelled)
         assert cancelled_in_context  # for the sleep_forever
 
     # Trying to exit a cancel scope from an unrelated task raises an error
@@ -1058,6 +846,7 @@ async def test_failed_abort():
     assert record == ["sleep", "woke", "cancelled"]
 
 
+@restore_unraisablehook()
 def test_broken_abort():
     async def main():
         # These yields are here to work around an annoying warning -- we're
@@ -1084,6 +873,7 @@ def test_broken_abort():
     gc_collect_harder()
 
 
+@restore_unraisablehook()
 def test_error_in_run_loop():
     # Blow stuff up real good to check we at least get a TrioInternalError
     async def main():
@@ -1255,13 +1045,18 @@ async def test_exc_info():
         nursery.start_soon(child2)
 
     assert record == [
-        "child1 raise", "child1 sleep", "child2 wake", "child2 sleep again",
-        "child1 re-raise", "child1 success", "child2 re-raise",
-        "child2 success"
+        "child1 raise",
+        "child1 sleep",
+        "child2 wake",
+        "child2 sleep again",
+        "child1 re-raise",
+        "child1 success",
+        "child2 re-raise",
+        "child2 success",
     ]
 
 
-# At least as of CPython 3.6, using .throw() to raise an exception inside a
+# Before CPython 3.9, using .throw() to raise an exception inside a
 # coroutine/generator causes the original exc_info state to be lost, so things
 # like re-raising and exception chaining are broken.
 #
@@ -1387,13 +1182,7 @@ async def test_TrioToken_run_sync_soon_idempotent():
         for i in range(100):
             token.run_sync_soon(cb, i, idempotent=True)
     await wait_all_tasks_blocked()
-    if (
-        sys.version_info < (3, 6)
-        and platform.python_implementation() == "CPython"
-    ):
-        # no order guarantees
-        record.sort()
-    # Otherwise, we guarantee FIFO
+    # We guarantee FIFO
     assert record == list(range(100))
 
 
@@ -1549,6 +1338,33 @@ async def test_TrioToken_run_sync_soon_massive_queue():
     assert counter[0] == COUNT
 
 
+@pytest.mark.skipif(buggy_pypy_asyncgens, reason="PyPy 7.2 is buggy")
+def test_TrioToken_run_sync_soon_late_crash():
+    # Crash after system nursery is closed -- easiest way to do that is
+    # from an async generator finalizer.
+    record = []
+    saved = []
+
+    async def agen():
+        token = _core.current_trio_token()
+        try:
+            yield 1
+        finally:
+            token.run_sync_soon(lambda: {}["nope"])
+            token.run_sync_soon(lambda: record.append("2nd ran"))
+
+    async def main():
+        saved.append(agen())
+        await saved[-1].asend(None)
+        record.append("main exiting")
+
+    with pytest.raises(_core.TrioInternalError) as excinfo:
+        _core.run(main)
+
+    assert type(excinfo.value.__cause__) is KeyError
+    assert record == ["main exiting", "2nd ran"]
+
+
 async def test_slow_abort_basic():
     with _core.CancelScope() as scope:
         scope.cancel()
@@ -1611,7 +1427,7 @@ async def test_task_tree_introspection():
     tasks = {}
     nurseries = {}
 
-    async def parent():
+    async def parent(task_status=_core.TASK_STATUS_IGNORED):
         tasks["parent"] = _core.current_task()
 
         assert tasks["parent"].child_nurseries == []
@@ -1624,7 +1440,7 @@ async def test_task_tree_introspection():
 
         async with _core.open_nursery() as nursery:
             nurseries["parent"] = nursery
-            nursery.start_soon(child1)
+            await nursery.start(child1)
 
         # Upward links survive after tasks/nurseries exit
         assert nurseries["parent"].parent_task is tasks["parent"]
@@ -1647,8 +1463,19 @@ async def test_task_tree_introspection():
         assert nurseries["child1"].child_tasks == frozenset({tasks["child2"]})
         assert tasks["child2"].child_nurseries == []
 
-    async def child1():
-        tasks["child1"] = _core.current_task()
+    async def child1(task_status=_core.TASK_STATUS_IGNORED):
+        me = tasks["child1"] = _core.current_task()
+        assert me.parent_nursery.parent_task is tasks["parent"]
+        assert me.parent_nursery is not nurseries["parent"]
+        assert me.eventual_parent_nursery is nurseries["parent"]
+        task_status.started()
+        assert me.parent_nursery is nurseries["parent"]
+        assert me.eventual_parent_nursery is None
+
+        # Wait for the start() call to return and close its internal nursery, to
+        # ensure consistent results in child2:
+        await _core.wait_all_tasks_blocked()
+
         async with _core.open_nursery() as nursery:
             nurseries["child1"] = nursery
             nursery.start_soon(child2)
@@ -1656,6 +1483,11 @@ async def test_task_tree_introspection():
     async with _core.open_nursery() as nursery:
         nursery.start_soon(parent)
 
+    # There are no pending starts, so no one should have a non-None
+    # eventual_parent_nursery
+    for task in tasks.values():
+        assert task.eventual_parent_nursery is None
+
 
 async def test_nursery_closure():
     async def child1(nursery):
@@ -1713,8 +1545,6 @@ async def test_current_effective_deadline(mock_clock):
     assert _core.current_effective_deadline() == inf
 
 
-# @coroutine is deprecated since python 3.8, which is fine with us.
-@pytest.mark.filterwarnings("ignore:.*@coroutine.*:DeprecationWarning")
 def test_nice_error_on_bad_calls_to_run_or_spawn():
     def bad_call_run(*args):
         _core.run(*args)
@@ -1726,66 +1556,26 @@ def test_nice_error_on_bad_calls_to_run_or_spawn():
 
         _core.run(main)
 
-    class Deferred:
-        "Just kidding"
+    for bad_call in bad_call_run, bad_call_spawn:
 
-    with ignore_coroutine_never_awaited_warnings():
-        for bad_call in bad_call_run, bad_call_spawn:
+        async def f():  # pragma: no cover
+            pass
 
-            async def f():  # pragma: no cover
-                pass
+        with pytest.raises(TypeError, match="expecting an async function"):
+            bad_call(f())
 
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(f())
-            assert "expecting an async function" in str(excinfo.value)
+        async def async_gen(arg):  # pragma: no cover
+            yield arg
 
-            import asyncio
-
-            @asyncio.coroutine
-            def generator_based_coro():  # pragma: no cover
-                yield from asyncio.sleep(1)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(generator_based_coro())
-            assert "asyncio" in str(excinfo.value)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(asyncio.Future())
-            assert "asyncio" in str(excinfo.value)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(lambda: asyncio.Future())
-            assert "asyncio" in str(excinfo.value)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(Deferred())
-            assert "twisted" in str(excinfo.value)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(lambda: Deferred())
-            assert "twisted" in str(excinfo.value)
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(len, [1, 2, 3])
-            assert "appears to be synchronous" in str(excinfo.value)
-
-            @async_generator
-            async def async_gen(arg):  # pragma: no cover
-                pass
-
-            with pytest.raises(TypeError) as excinfo:
-                bad_call(async_gen, 0)
-            msg = "expected an async function but got an async generator"
-            assert msg in str(excinfo.value)
-
-            # Make sure no references are kept around to keep anything alive
-            del excinfo
+        with pytest.raises(
+            TypeError, match="expected an async function but got an async generator"
+        ):
+            bad_call(async_gen, 0)
 
 
 def test_calling_asyncio_function_gives_nice_error():
     async def child_xyzzy():
-        import asyncio
-        await asyncio.Future()
+        await create_asyncio_future_in_new_loop()
 
     async def misguided():
         await child_xyzzy()
@@ -1805,8 +1595,9 @@ async def test_asyncio_function_inside_nursery_does_not_explode():
     with pytest.raises(TypeError) as excinfo:
         async with _core.open_nursery() as nursery:
             import asyncio
+
             nursery.start_soon(sleep_forever)
-            await asyncio.Future()
+            await create_asyncio_future_in_new_loop()
     assert "asyncio" in str(excinfo.value)
 
 
@@ -1828,10 +1619,10 @@ async def test_trivial_yields():
             async with _core.open_nursery():
                 raise KeyError
         assert len(excinfo.value.exceptions) == 2
-        assert {type(e)
-                for e in excinfo.value.exceptions} == {
-                    KeyError, _core.Cancelled
-                }
+        assert {type(e) for e in excinfo.value.exceptions} == {
+            KeyError,
+            _core.Cancelled,
+        }
 
 
 async def test_nursery_start(autojump_clock):
@@ -1843,9 +1634,7 @@ async def test_nursery_start(autojump_clock):
         with pytest.raises(TypeError):
             await nursery.start(no_args)
 
-    async def sleep_then_start(
-        seconds, *, task_status=_core.TASK_STATUS_IGNORED
-    ):
+    async def sleep_then_start(seconds, *, task_status=_core.TASK_STATUS_IGNORED):
         repr(task_status)  # smoke test
         await sleep(seconds)
         task_status.started(seconds)
@@ -1909,9 +1698,7 @@ async def test_nursery_start(autojump_clock):
 
     # and if after the no-op started(), the child crashes, the error comes out
     # of start()
-    async def raise_keyerror_after_started(
-        task_status=_core.TASK_STATUS_IGNORED
-    ):
+    async def raise_keyerror_after_started(task_status=_core.TASK_STATUS_IGNORED):
         task_status.started()
         raise KeyError("whoopsiedaisy")
 
@@ -1920,8 +1707,10 @@ async def test_nursery_start(autojump_clock):
             cs.cancel()
             with pytest.raises(_core.MultiError) as excinfo:
                 await nursery.start(raise_keyerror_after_started)
-    assert {type(e)
-            for e in excinfo.value.exceptions} == {_core.Cancelled, KeyError}
+    assert {type(e) for e in excinfo.value.exceptions} == {
+        _core.Cancelled,
+        KeyError,
+    }
 
     # trying to start in a closed nursery raises an error immediately
     async with _core.open_nursery() as closed_nursery:
@@ -2056,15 +1845,12 @@ async def test_nursery_stop_async_iteration():
         async def _accumulate(self, f, items, i):
             items[i] = await f()
 
-        @aiter_compat
         def __aiter__(self):
             return self
 
         async def __anext__(self):
             nexts = self.nexts
-            items = [
-                None,
-            ] * len(nexts)
+            items = [None] * len(nexts)
             got_stop = False
 
             def handle(exc):
@@ -2154,7 +1940,7 @@ async def test_contextvar_multitask():
 
 
 def test_system_task_contexts():
-    cvar = contextvars.ContextVar('qwilfish')
+    cvar = contextvars.ContextVar("qwilfish")
     cvar.set("water")
 
     async def system_task():
@@ -2174,11 +1960,7 @@ def test_system_task_contexts():
 
 
 def test_Nursery_init():
-    check_Nursery_error = pytest.raises(
-        TypeError, match='no public constructor available'
-    )
-
-    with check_Nursery_error:
+    with pytest.raises(TypeError):
         _core._run.Nursery(None, None)
 
 
@@ -2189,23 +1971,17 @@ async def test_Nursery_private_init():
 
 
 def test_Nursery_subclass():
-    with pytest.raises(
-        TypeError, match='`Nursery` does not support subclassing'
-    ):
+    with pytest.raises(TypeError):
 
         class Subclass(_core._run.Nursery):
             pass
 
 
 def test_Cancelled_init():
-    check_Cancelled_error = pytest.raises(
-        TypeError, match='no public constructor available'
-    )
-
-    with check_Cancelled_error:
+    with pytest.raises(TypeError):
         raise _core.Cancelled
 
-    with check_Cancelled_error:
+    with pytest.raises(TypeError):
         _core.Cancelled()
 
     # private constructor should not raise
@@ -2214,22 +1990,18 @@ def test_Cancelled_init():
 
 def test_Cancelled_str():
     cancelled = _core.Cancelled._create()
-    assert str(cancelled) == 'Cancelled'
+    assert str(cancelled) == "Cancelled"
 
 
 def test_Cancelled_subclass():
-    with pytest.raises(
-        TypeError, match='`Cancelled` does not support subclassing'
-    ):
+    with pytest.raises(TypeError):
 
         class Subclass(_core.Cancelled):
             pass
 
 
 def test_CancelScope_subclass():
-    with pytest.raises(
-        TypeError, match='`CancelScope` does not support subclassing'
-    ):
+    with pytest.raises(TypeError):
 
         class Subclass(_core.CancelScope):
             pass
@@ -2276,9 +2048,7 @@ async def test_permanently_detach_coroutine_object():
         await async_yield(yield_value)
 
     async with _core.open_nursery() as nursery:
-        nursery.start_soon(
-            detachable_coroutine, outcome.Value(None), "I'm free!"
-        )
+        nursery.start_soon(detachable_coroutine, outcome.Value(None), "I'm free!")
 
     # If we get here then Trio thinks the task has exited... but the coroutine
     # is still iterable
@@ -2293,9 +2063,7 @@ async def test_permanently_detach_coroutine_object():
     pdco_outcome = None
     with pytest.raises(KeyError):
         async with _core.open_nursery() as nursery:
-            nursery.start_soon(
-                detachable_coroutine, outcome.Error(KeyError()), "uh oh"
-            )
+            nursery.start_soon(detachable_coroutine, outcome.Error(KeyError()), "uh oh")
     throw_in = ValueError()
     assert task.coro.throw(throw_in) == "uh oh"
     assert pdco_outcome == outcome.Error(throw_in)
@@ -2305,9 +2073,7 @@ async def test_permanently_detach_coroutine_object():
     async def bad_detach():
         async with _core.open_nursery():
             with pytest.raises(RuntimeError) as excinfo:
-                await _core.permanently_detach_coroutine_object(
-                    outcome.Value(None)
-                )
+                await _core.permanently_detach_coroutine_object(outcome.Value(None))
             assert "open nurser" in str(excinfo.value)
 
     async with _core.open_nursery() as nursery:
@@ -2338,9 +2104,7 @@ async def test_detach_and_reattach_coroutine_object():
         await async_yield(2)
 
         with pytest.raises(RuntimeError) as excinfo:
-            await _core.reattach_detached_coroutine_object(
-                unrelated_task, None
-            )
+            await _core.reattach_detached_coroutine_object(unrelated_task, None)
         assert "does not match" in str(excinfo.value)
 
         await _core.reattach_detached_coroutine_object(task, "byebye")
@@ -2392,27 +2156,15 @@ async def test_detached_coroutine_cancellation():
     assert abort_fn_called
 
 
+@restore_unraisablehook()
 def test_async_function_implemented_in_C():
     # These used to crash because we'd try to mutate the coroutine object's
     # cr_frame, but C functions don't have Python frames.
 
-    ns = {"_core": _core}
-    try:
-        exec(
-            dedent(
-                """
-                async def agen_fn(record):
-                    assert not _core.currently_ki_protected()
-                    record.append("the generator ran")
-                    yield
-                """
-            ),
-            ns,
-        )
-    except SyntaxError:
-        pytest.skip("Requires Python 3.6+")
-    else:
-        agen_fn = ns["agen_fn"]
+    async def agen_fn(record):
+        assert not _core.currently_ki_protected()
+        record.append("the generator ran")
+        yield
 
     run_record = []
     agen = agen_fn(run_record)
@@ -2437,3 +2189,145 @@ async def test_very_deep_cancel_scope_nesting():
         for _ in range(5000):
             exit_stack.enter_context(_core.CancelScope())
         outermost_scope.cancel()
+
+
+async def test_cancel_scope_deadline_duplicates():
+    # This exercises an assert in Deadlines._prune, by intentionally creating
+    # duplicate entries in the deadline heap.
+    now = _core.current_time()
+    with _core.CancelScope() as cscope:
+        for _ in range(DEADLINE_HEAP_MIN_PRUNE_THRESHOLD * 2):
+            cscope.deadline = now + 9998
+            cscope.deadline = now + 9999
+        await sleep(0.01)
+
+
+@pytest.mark.skipif(
+    sys.implementation.name != "cpython", reason="Only makes sense with refcounting GC"
+)
+async def test_simple_cancel_scope_usage_doesnt_create_cyclic_garbage():
+    # https://github.com/python-trio/trio/issues/1770
+    gc.collect()
+
+    async def do_a_cancel():
+        with _core.CancelScope() as cscope:
+            cscope.cancel()
+            await sleep_forever()
+
+    async def crasher():
+        raise ValueError
+
+    old_flags = gc.get_debug()
+    try:
+        gc.collect()
+        gc.set_debug(gc.DEBUG_SAVEALL)
+
+        # cover outcome.Error.unwrap
+        # (See https://github.com/python-trio/outcome/pull/29)
+        await do_a_cancel()
+        # cover outcome.Error.unwrap if unrolled_run hangs on to exception refs
+        # (See https://github.com/python-trio/trio/pull/1864)
+        await do_a_cancel()
+
+        with pytest.raises(ValueError):
+            async with _core.open_nursery() as nursery:
+                # cover MultiError.filter and NurseryManager.__aexit__
+                nursery.start_soon(crasher)
+
+        gc.collect()
+        assert not gc.garbage
+    finally:
+        gc.set_debug(old_flags)
+        gc.garbage.clear()
+
+
+@pytest.mark.skipif(
+    sys.implementation.name != "cpython", reason="Only makes sense with refcounting GC"
+)
+async def test_cancel_scope_exit_doesnt_create_cyclic_garbage():
+    # https://github.com/python-trio/trio/pull/2063
+    gc.collect()
+
+    async def crasher():
+        raise ValueError
+
+    old_flags = gc.get_debug()
+    try:
+
+        with pytest.raises(ValueError), _core.CancelScope() as outer:
+            async with _core.open_nursery() as nursery:
+                gc.collect()
+                gc.set_debug(gc.DEBUG_SAVEALL)
+                # One child that gets cancelled by the outer scope
+                nursery.start_soon(sleep_forever)
+                outer.cancel()
+                # And one that raises a different error
+                nursery.start_soon(crasher)
+                # so that outer filters a Cancelled from the MultiError and
+                # covers CancelScope.__exit__ (and NurseryManager.__aexit__)
+                # (See https://github.com/python-trio/trio/pull/2063)
+
+        gc.collect()
+        assert not gc.garbage
+    finally:
+        gc.set_debug(old_flags)
+        gc.garbage.clear()
+
+
+@pytest.mark.skipif(
+    sys.implementation.name != "cpython", reason="Only makes sense with refcounting GC"
+)
+async def test_nursery_cancel_doesnt_create_cyclic_garbage():
+    # https://github.com/python-trio/trio/issues/1770#issuecomment-730229423
+    def toggle_collected():
+        nonlocal collected
+        collected = True
+
+    collected = False
+    gc.collect()
+    old_flags = gc.get_debug()
+    try:
+        gc.set_debug(0)
+        gc.collect()
+        gc.set_debug(gc.DEBUG_SAVEALL)
+
+        # cover Nursery._nested_child_finished
+        async with _core.open_nursery() as nursery:
+            nursery.cancel_scope.cancel()
+
+        weakref.finalize(nursery, toggle_collected)
+        del nursery
+        # a checkpoint clears the nursery from the internals, apparently
+        # TODO: stop event loop from hanging on to the nursery at this point
+        await _core.checkpoint()
+
+        assert collected
+        gc.collect()
+        assert not gc.garbage
+    finally:
+        gc.set_debug(old_flags)
+        gc.garbage.clear()
+
+
+@pytest.mark.skipif(
+    sys.implementation.name != "cpython", reason="Only makes sense with refcounting GC"
+)
+async def test_locals_destroyed_promptly_on_cancel():
+    destroyed = False
+
+    def finalizer():
+        nonlocal destroyed
+        destroyed = True
+
+    class A:
+        pass
+
+    async def task():
+        a = A()
+        weakref.finalize(a, finalizer)
+        await _core.checkpoint()
+
+    async with _core.open_nursery() as nursery:
+        nursery.start_soon(task)
+        nursery.cancel_scope.cancel()
+    assert destroyed
